#include <time.h>
#include "graph.h"

#ifndef NT
#define NT 1000
#endif

static LSUP_Triple *
generate_triples()
{
    LSUP_Triple *trp;
    trp = malloc((NT  + 1) * sizeof(LSUP_Triple));
    if (!trp) exit (-1);

    for (size_t i = 0; i < NT; i++) {
        char sstr[32], pstr[32], ostr[32];

        sprintf(sstr, "urn:s:%lu", i % (NT / 100));
        sprintf(pstr, "urn:p:%lu", i % (NT / 1000));
        sprintf(ostr, "urn:o:%lu", i);
        LSUP_triple_init(
                trp + i, LSUP_uri_new (sstr),
                LSUP_uri_new (pstr), LSUP_uri_new (ostr));
    }
    LSUP_triple_init (trp + NT, NULL, NULL, NULL);
    log_info ("Triples generated.");

    return trp;
}

static LSUP_rc
insert_triples (LSUP_Graph *gr, LSUP_Triple *trp)
{
    size_t ct;
    LSUP_rc rc = LSUP_graph_add_trp(gr, trp, &ct);
    if (rc != LSUP_OK) log_warn ("Graph loading interrupted: %d.", rc);
    else log_info ("Graph populated with %lu triples.", ct);

    return rc;
}


int main()
{
    // Set env variable to test path.
    putenv ("LSUP_MDB_STORE_PATH=" TMPDIR "/lsup_profile_mdb");
    // Clear out database from previous test.
    rm_r (getenv ("LSUP_MDB_STORE_PATH"));
    LSUP_init();

    int rc;
    clock_t start, tc1, tc2, end;
    double wallclock, rate;

    log_info ("Generating triples.");
    start = clock();
    LSUP_Triple *trp = generate_triples();
    tc1 = clock();
    wallclock = (tc1 - start) / CLOCKS_PER_SEC;
    log_info ("Time elapsed: %lf s", wallclock);

    log_info ("Inserting triples.");
    LSUP_Graph *gr = LSUP_graph_new (LSUP_STORE_MDB);
    if (!gr) {
        log_error ("Error creating graph!");
        return -1;
    }
    rc = insert_triples (gr, trp);

    for (size_t i = 0; i < NT; i++) {
        LSUP_term_free (trp[i].s);
        LSUP_term_free (trp[i].p);
        LSUP_term_free (trp[i].o);
    }
    free (trp);

    tc2 = clock();
    wallclock = (tc2 - tc1) / CLOCKS_PER_SEC;
    log_info ("Time elapsed: %lf s", wallclock);
    log_info ("Graph size: %lu", LSUP_graph_size (gr));

    log_info ("Lookup...");
    size_t ct = 0;
    LSUP_Triple *spo = TRP_DUMMY;
    LSUP_Term *s = LSUP_uri_new ("urn:s:0");
    LSUP_Term *p = LSUP_uri_new ("urn:p:3200");
    LSUP_Term *o = LSUP_uri_new ("urn:o:3200");
    LSUP_GraphIterator *it = LSUP_graph_lookup(gr, NULL, NULL, NULL, NULL);
    while (LSUP_graph_iter_next (it, spo) != LSUP_END)
        ct ++;
    log_info ("Found triples per subject: %lu", ct);
    LSUP_graph_iter_free (it);
    end = clock();
    wallclock = (end - tc2) / CLOCKS_PER_SEC;
    log_info ("Time elapsed: %lf s", wallclock);

    wallclock = (end - start) / CLOCKS_PER_SEC;
    rate = NT / wallclock;
    log_info (
            "%d triples created and inserted in %lf s (%lf triples/s)",
            NT, wallclock, rate);

    LSUP_term_free (s);
    LSUP_term_free (p);
    LSUP_term_free (o);
    LSUP_graph_free(gr);

    return rc;
}
